"""
Type-safe (ish) YAML configuration and validation.

An example ``config.py`` file:

```python
from typing import Optional

import yaml

from configur8 import cfg, env


class MySQLConfig:
    host: str
    port: int
    username: Optional[str]
    password: Optional[str]
    database: str

    def __init__(
        self,
        # shame that nested inference is not a thing in MyPy yet
        host: str,
        port: int,
        username: Optional[str],
        password: Optional[str],
        database: str,
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database


class Config:
    mysql: MySQLConfig

    def __init__(self, mysql: MySQLConfig) -> None:
        self.mysql = mysql


def get_mysql(data: cfg.YamlConfig) -> MySQLConfig:
    return MySQLConfig(
        host=data.str("host"),
        port=data.int("port"),
        username=data.str_optional("username"),
        password=data.str_optional("password"),
        database=data.str("database"),
    )


def parse(data: str) -> Config:
    config = yaml.safe_load(data)

    if not isinstance(config, dict):
        raise cfg.ConfigError("", "Expected root from yaml to be a dict")

    config = cfg.YamlConfig(config)

    return Config(
        mysql=get_mysql(config.with_prefix("mysql")),
    )


def load(path: Optional[str] = None) -> Config:
    if path is None:
        # Load from environment
        path = env.str("CONFIG_PATH")

    with open(path, "rb") as fp:
        return parse(fp.read().decode("utf-8"))
```
"""

import builtins
import re
from typing import Any, Dict, List, Optional, Sequence, Type, TypeVar, Union


class Missing:
    pass


MISSING = Missing()
Data = TypeVar("Data")
PathLike = Sequence[Union[str, int]]


class Path:
    data: List[Union[str, int]]

    match = re.compile(r"([a-zA-Z0-9_-]+)\.?|\[([0-9]+)\]\.?")

    def __init__(self, path: PathLike) -> None:
        self.data = list(path)

    def __str__(self) -> str:
        ret = ""

        first = True

        for count, part in enumerate(self.data):
            if isinstance(part, str):
                if not first:
                    ret += "."

                ret += part
            elif isinstance(part, int):
                ret += f"[{part}]"
            else:
                raise TypeError(
                    f"Unexpected {type(part)} at {count} in {self.path!r}"
                )

            first = False

        return ret

    def __iter__(self):
        return iter(self.data)

    @staticmethod
    def decode(path: str) -> "Path":
        ret: List[Union[str, int]] = []

        for s, i in Path.match.findall(path):
            if len(s) > 0:
                try:
                    ret.append(int(s))
                except ValueError:
                    ret.append(s)
            elif len(i) > 0:
                ret.append(int(i))
            else:
                raise RuntimeError

        return Path(ret)

    def __add__(self, other: Any) -> "Path":
        if isinstance(other, str):
            return Path(self.data + Path.decode(other).data)
        elif isinstance(other, Path):
            return Path(self.data + other.data)
        else:
            raise TypeError(f"{other!r}<{type(other)}>")


class ConfigError(Exception):
    path: Path

    def __init__(
        self,
        path: Union[PathLike, Path, str],
        message: str,
    ) -> None:
        self.message = message

        if isinstance(path, Path):
            self.path = path
        elif isinstance(path, str):
            self.path = Path.decode(path)
        else:
            self.path = Path(path)

    def __str__(self) -> str:
        return f"{self.path}: {self.message}"


class YamlConfig:
    root: Dict[str, Any]
    prefix: Optional[Path]

    def __init__(
        self,
        root: Dict[str, Any],
        prefix: Optional[Union[str, Path]] = None,
    ) -> None:
        self.root = root

        if prefix is None:
            self.prefix = prefix
        else:
            if isinstance(prefix, str):
                self.prefix = Path.decode(prefix)
            else:
                self.prefix = prefix

    def path(
        self,
        path: str,
        default: Union[Missing, Any] = MISSING,
    ) -> Any:
        obj = self.root

        if self.prefix is not None:
            path = str(self.prefix) + "." + path

        for part in Path.decode(path):
            try:
                obj = obj[part]
            except KeyError as err:
                if default is not MISSING:
                    return default

                raise ConfigError(path.split("."), "missing") from err
            except TypeError as err:
                # report this please
                print(Path.decode(path).data)
                print(path, type(part), obj)

                raise ConfigError(path.split("."), "invalid") from err

        return obj

    def get(
        self,
        type_: Type[Data],
        path: str,
        default: Union[Missing, Data] = MISSING,
    ) -> Data:
        value = self.path(path, default)

        if value == default:
            if not isinstance(value, type_):
                raise TypeError(
                    f"Expected {type_} at {path!r}, got {value!r} instead"
                )

            return value

        if not isinstance(value, type_):
            raise ConfigError(
                self.with_path(path),
                f"{type_} expected but found {type(value)}"
            )

        return value

    def optional(self, type_: Type[Data], path: str) -> Optional[Data]:
        try:
            value = self.path(path)
        except ConfigError:
            return None

        if not isinstance(value, type_):
            raise ConfigError(
                self.with_path(path),
                f"{type_} expected but found {type(value)}"
            )

        return value

    def with_path(self, path: str) -> str:
        if self.prefix is None:
            return path

        return str(self.prefix + path)

    def with_prefix(self, prefix: str) -> "YamlConfig":
        if self.prefix is None:
            return YamlConfig(self.root, prefix)

        return YamlConfig(self.root, f"{self.prefix}.{prefix}")

    def str(
        self,
        path: str,
        default: Union[Missing, str] = MISSING,
    ) -> str:
        return self.get(str, path, default)

    def str_optional(self, path: builtins.str) -> Optional[builtins.str]:
        return self.optional(str, path)

    def int(
        self,
        path: builtins.str,
        default: Union[Missing, int] = MISSING,
    ) -> int:
        return self.get(int, path, default)

    def int_optional(self, path: builtins.str) -> Optional[builtins.int]:
        return self.optional(int, path)

    def list(
        self,
        path: builtins.str,
        default: Union[Missing, list] = MISSING,
    ) -> List[Any]:
        return self.get(list, path, default)

    def list_optional(self, path: builtins.str) -> Optional[List[Any]]:
        return self.optional(list, path)

    def dict(
        self,
        path: builtins.str,
        default: Union[Missing, builtins.dict] = MISSING,
    ) -> Dict[builtins.str, Any]:
        return self.get(dict, path, default)

    def dict_optional(
        self,
        path: builtins.str
    ) -> Optional[Dict[builtins.str, Any]]:
        return self.optional(dict, path)



from email_validator import EmailNotValidError, validate_email

from .core import InvalidConfig

__all__ = ("parse",)


def parse(email: str) -> str:
    try:
        result = validate_email(email, check_deliverability=False)
    except EmailNotValidError as exc:
        raise InvalidConfig(str(exc)) from exc

    if not result.email:
        raise InvalidConfig(f"Email is not valid?? ({email!r})")

    assert isinstance(result.email, str)

    return result.email



from configur8.__about__ import version as __version__  # noqa: F401
from configur8.__about__ import version_info as __version_info__  # noqa: F401

from .core import InvalidConfig
from .path import Path
from .url import Url

__all__ = (
    "InvalidConfig",
    "Url",
    "Path",
)



from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import ParseResult
from urllib.parse import urlparse

__all__ = (
    "parse",
    "Url",
)


class Url:
    result: ParseResult

    def __init__(self, result: ParseResult):
        self.result = result

    def __str__(self):
        return self.result.geturl()

    def __repr__(self):
        return f"<{__name__}.{self.__class__.__name__} {str(self)}>"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Url):
            return self.result == other.result

        if isinstance(other, str):
            return str(self) == other

        raise TypeError(f"Cannot compare {other!r} with <Url>")

    @property
    def protocol(self) -> Optional[str]:
        if self.result.scheme == "":
            return None

        return self.result.scheme

    @property
    def username(self) -> Optional[str]:
        return self.result.username

    @property
    def password(self) -> Optional[str]:
        return self.result.password

    @property
    def host(self) -> str:
        netloc = self.result.netloc

        if netloc == "":
            return self.result.path.split("/")[0]

        try:
            host = netloc.split("@")[1]
        except IndexError:
            # no username/password
            return netloc.split(":")[0]

        return host.split(":")[0]

    @property
    def port(self) -> Optional[int]:
        return self.result.port

    @property
    def path(self) -> Optional[str]:
        path = self.result.path

        if self.result.netloc == "":
            path = "/".join(self.result.path.split("/")[1:])

            if path:
                path = "/" + path

        if path == "":
            return None

        return path

    @property
    def query(self) -> Optional[Dict[str, List[str]]]:
        if self.result.query == "":
            return None

        return parse_qs(
            self.result.query,
            keep_blank_values=True,
            strict_parsing=True,
        )

    @property
    def fragment(self) -> Optional[str]:
        if self.result.fragment == "":
            return None

        return self.result.fragment

    # support the str methods
    def __getattr__(self, name):
        return getattr(str(self), name)


def parse(data: Union[str, Url]) -> Url:
    if isinstance(data, Url):
        return data

    return Url(urlparse(data))



from typing import List, Optional
from typing import Union

__all__ = (
    "parse",
    "Path",
)


class Path:
    path: str

    def __init__(self, path: str):
        self.path = path

    def __str__(self):
        return self.path

    def __repr__(self) -> str:
        return f"<{__name__}.{self.__class__.__name__} {str(self)}>"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Path):
            return self.path == other.path

        if isinstance(other, str):
            return str(self) == other

        raise TypeError(f"Cannot compare {other!r} with <Url>")

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
    ):
        """
        See https://docs.python.org/3/library/functions.html#open for arg
        documentation
        """
        return open(
            self.path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )

    def read(self) -> str:
        ret = self.open(mode="rt", encoding="utf-8").read()

        return ret  # type: ignore

    def readlines(self, newline: Optional[str] = None) -> List[str]:
        ret = self.open(newline=newline).readlines()

        return ret  # type: ignore


def parse(path: Union[str, Path]) -> Path:
    if isinstance(path, Path):
        return path

    return Path(path)



"""
Helper to pull configuration values from the environment.

Example:

```python
from configur8 import env

SECRET_KEY = env.str("SECRET_KEY")
NUM_WORKERS = env.int("NUM_WORKERS", 2)
DEBUG = env.bool("DEBUG", False)
```

In the example above:
1. ``SECRET_KEY`` is a required environment variable - attempting execute the
   code without it defined will result in an exception. This is typically what
   you want so that apps and services don't start in an unintended state.
2. ``NUM_WORKERS`` will be parsed into an integer. If the env var is not
   defined, ``2`` will be used as a default. If a non integer value is parsed,
   an error will be raised.
3. ``DEBUG`` is a boolean with a default of ``False``. "Truthy" values can be
   used, e.g. "on", "1", etc.

Everything is designed to be type safe.
"""
import builtins
import os
from typing import Callable
from typing import Generic
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from .core import InvalidConfig
from .email import parse as parse_email
from .path import parse as parse_path
from .path import Path
from .url import parse as parse_url
from .url import Url
from .util import Empty
from .util import MISSING

__all__ = (
    "MissingFromEnv",
    "str",
    "bool",
    "int",
    "float",
    "url",
    "path",
    "email",
)

T = TypeVar("T")

BOOLEAN_TRUTHY_VALUES = ["true", "1", "y", "yes", "ok"]
LIST_SEPARATOR = ","


class MissingFromEnv(InvalidConfig):
    """
    Raised if an attempt to get an environment variable fails
    """


def get_raw(env_var_name: builtins.str) -> builtins.str:
    try:
        return os.environ[env_var_name]
    except KeyError:
        raise MissingFromEnv(f"Missing env var {env_var_name!r}")


def get_raw_optional(env_var: builtins.str) -> Optional[builtins.str]:
    return os.getenv(env_var, None)


class EnvVar(Generic[T]):
    def __init__(self, parse_func: Callable[[Union[builtins.str, T]], T]):
        self.parse_func = parse_func

    def default(
        self,
        env_var_name: builtins.str,
        default: Union[Empty, builtins.str, T, List[T]] = MISSING,
    ) -> Union[builtins.str, T, List[T]]:
        if default is MISSING:
            return get_raw(env_var_name)

        if TYPE_CHECKING:
            assert not isinstance(default, Empty)

        raw_value = get_raw_optional(env_var_name)

        if raw_value is None:
            return default

        return raw_value

    def __call__(
        self,
        env_var_name: builtins.str,
        default: Union[Empty, builtins.str, T] = MISSING,
    ) -> T:
        raw_value = self.default(env_var_name, default)

        if TYPE_CHECKING:
            assert not isinstance(raw_value, Empty)
            assert not isinstance(raw_value, list)

        return self.parse_func(raw_value)

    def optional(self, env_var_name: builtins.str) -> Optional[T]:
        raw_value = get_raw_optional(env_var_name)

        if raw_value is None:
            return None

        return self.parse_func(raw_value)

    def list(
        self,
        env_var_name: builtins.str,
        default: Union[Empty, builtins.str, List[T]] = MISSING,
        separator=LIST_SEPARATOR,
    ) -> List[T]:
        raw_value = self.default(env_var_name, default)

        if isinstance(raw_value, list):
            return raw_value

        if TYPE_CHECKING:
            assert not isinstance(raw_value, Empty)
            assert isinstance(raw_value, builtins.str)

        return [self.parse_func(item) for item in raw_value.split(separator)]

    def list_optional(
        self,
        env_var_name: builtins.str,
        separator=LIST_SEPARATOR,
    ) -> Optional[List[T]]:
        raw_value = get_raw_optional(env_var_name)

        if raw_value is None:
            return None

        return [self.parse_func(item) for item in raw_value.split(separator)]


def parse_str(raw_value: builtins.str) -> builtins.str:
    return raw_value


def parse_bool(raw_value: Union[builtins.str, builtins.bool]) -> builtins.bool:
    if isinstance(raw_value, builtins.bool):
        return raw_value

    return raw_value.lower() in BOOLEAN_TRUTHY_VALUES


def parse_int(raw_value: Union[builtins.str, builtins.int]) -> builtins.int:
    if isinstance(raw_value, builtins.int):
        return raw_value

    try:
        return builtins.int(raw_value)
    except ValueError:
        raise InvalidConfig(f"{raw_value!r} is not a valid integer")


def parse_float(
    raw_value: Union[builtins.str, builtins.float],
) -> builtins.float:
    try:
        return builtins.float(raw_value)
    except ValueError:
        raise InvalidConfig(f"{raw_value!r} is not a valid number")


str = EnvVar[builtins.str](parse_str)
bool = EnvVar[builtins.bool](parse_bool)
int = EnvVar[builtins.int](parse_int)
float = EnvVar[builtins.float](parse_float)
url = EnvVar[Url](parse_url)
path = EnvVar[Path](parse_path)
email = EnvVar[builtins.str](parse_email)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['configur8']

package_data = \
{'': ['*']}

install_requires = \
['email-validator>=1,<2']

setup_kwargs = {
    'name': 'configur8',
    'version': '1.6',
    'description': 'Type-safe configuration and validation library',
    'long_description': 'None',
    'author': 'Nick Joyce',
    'author_email': 'nick@stratuscode.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
